@@ -2,7 +2,7 @@
 
 // 2007-03-12  Stephen M. Webb  <stephen.webb@bregmasoft.com>
 //
-// Copyright (C) 2007-2014 Free Software Foundation, Inc.
+// Copyright (C) 2007-2019 Free Software Foundation, Inc.
 //
 // This file is part of the GNU ISO C++ Library.  This library is free
 // software; you can redistribute it and/or modify it under the
@@ -28,14 +28,13 @@
 // default constructor.
 void test01()
 {
-  bool test __attribute__((unused)) = true;
   std::tr1::regex_traits<char> t;
   VERIFY( t.translate_nocase('A') == 'a' );
 }
 
 int
 main()
-{ 
+{
   test01();
   return 0;
 }
